# Copyright (c) 2006 by BlueDynamics Tyrol - Klein und Partner KEG, Austria
#
# GNU General Public License (GPL)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__author__ = """Jens Klein <jens@bluedynamics.com>"""
__docformat__ = 'plaintext'

import socket, os
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from Globals import InitializeClass
from Globals import package_home
from BTrees.OOBTree import OOBTree
from OFS.Cache import Cacheable
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluggableAuthService.utils import createViewName
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import IGroupsPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.permissions import ManageGroups
from Products.PASGroupByIP import ipaddr

_wwwdir = os.path.join( package_home( globals() ), 'www' )

manage_addGroupByIPPluginForm = PageTemplateFile(
    os.path.join(_wwwdir, 'addGroupByIPPlugin'), 
    globals(),
    __name__='manage_addGroupByIPPlugin')

def addGroupByIPPlugin(self, id, title='', REQUEST=None):
    """Factory method to instantiate a addGroupByIPPlugin."""
    
    # Make sure we really are working in our container (the 
    # PluggableAuthService object)
    self = self.this()

    # Instantiate the folderish adapter object
    plugin = GroupByIPPlugin(id, title=title)
    self._setObject(id, plugin)

    if REQUEST is not None:
        REQUEST.RESPONSE.redirect('%s/manage_main' % self.absolute_url())
        
class GroupByIPPlugin(BasePlugin, Cacheable):
    ''' make principals part of a group because of their IP-address '''

    security = ClassSecurityInfo()
    
    __implements__ = (getattr(BasePlugin,'__implements__',()),)
    
    meta_type = 'GroupByIPPlugin'
        
    def __init__(self, id, title=None):
        self.id = id
        self.title = title
        self._groupipmap = OOBTree()        

    def getGroupsForPrincipal(self, principal, request=None ):

        """ principal -> ( group_1, ... group_N )

        o Return a sequence of group names to which the principal 
          (either a user or another group) belongs.

        """
        if request is None:
            request = self.REQUEST    
        ip = self._extractIP(request)
        if ip is not None:
            return self._findGroups(ip)
        else:
            return tuple()
        
        
    def _extractIP(self, request):
        """Extracts the IP from the REQUEST.
        
        This is the weak part.
        """
        ip = request.get( 'HTTP_X_FORWARDED_FOR', None ) 
        if ip:
            ip = ip.split(',')[0].strip()
            return ip

        # fetch ip from REQUEST
        if hasattr(request, 'getClientAddr'):
            ip = request.getClientAddr()
            return ip
            
        ip = request.get( 'REMOTE_ADDR', None ) 
        if ip:
            return ip
            
        hostname = request.get( 'REMOTE_HOST', '' )
        if hostname:
            try:
                ip = socket.gethostbyname(hostname)
            except:
                pass
            else:
                return ip           
        return None


    def _findGroups(self, ip):
        """ returns a list of groups the ip matches """
        try:
            ip = ipaddr.ipaddr(ip)
        except:
            return tuple()
        view_name = createViewName('PASGroupByIPPlugin', self.id)
        keywords = {'key': str(ip)}
        cached = self.ZCacheable_get(view_name=view_name,
                                     keywords=keywords,
                                     default=None)        
        if cached is not None:
            return tuple(cached)
        result = []
        for group in self._groupipmap:
            if ip in self._groupipmap[group]:
                result.append(group)
        self.ZCacheable_set(result, view_name=view_name, keywords=keywords)
        return tuple(result)

    
    security.declarePrivate( 'addIPRangeToGroup' )
    def addIPRangeToGroup(self, group, ip, mask=None):
        """ adds an iprange given by ip and netmask to a group"""
        if not self._groupipmap.has_key(group):
            self._groupipmap[group] = ipaddr.ip_set()
        if mask is None:
            nwip = ipaddr.ipaddr(ip)
        else:
            nwip = ipaddr.network(ip, mask)
        view_name = createViewName('PASGroupByIPPlugin', self.id)
        self.ZCacheable_invalidate(view_name=view_name)
        self._groupipmap[group].append(nwip)


    security.declarePrivate( 'removeIPRangeFromGroup' )
    def removeIPRangeFromGroup(self, group, ip, mask=None):
        """ adds an iprange given by ip and netmask to a group"""
        if not self._groupipmap.has_key(group):
            return
        if mask is None:
            nwip = ipaddr.ipaddr(ip)
            self._groupipmap[group].subtract_ips([nwip])
        else:
            nwip = ipaddr.network(ip, mask)
            self._groupipmap[group].subtract_network(nwip)
        view_name = createViewName('PASGroupByIPPlugin', self.id)
        self.ZCacheable_invalidate(view_name=view_name)

    security.declareProtected( ManageGroups, 'listIPManagedGroups' )
    def listIPManagedGroups(self):
        """ returns a list of Groups managed by this plugin """
        return self._groupipmap.keys()        
    
    security.declareProtected( ManageGroups, 'listIPRangesOfGroup' )
    def listIPRangesOfGroup(self, group):
        """ returns a list of all IP Ranges of a certain group """
        if not self._groupipmap.has_key(group):
            return None
        result = [str(n) for n in self._groupipmap[group].as_list_of_networks()]
        result.sort()
        return result

    security.declareProtected( ManageGroups, 'manage_changeIPRangeGroupMapping' )
    def manage_changeIPRangeGroupMapping(self, REQUEST=None):
        """ manages submitted form """
        added = {}
        removed = {}
        for key in REQUEST.form.keys():
            if key.startswith('ADDRANGE_'):
                group = key[9:]
                values = REQUEST.form[key].split(';')
                for value in values:
                    value = value.strip()
                    if value:
                        mask = None
                        val_mask = value.split('/')
                        if len(val_mask)==2 and val_mask[1]:
                            mask = val_mask[1]
                            value = val_mask[0]
                        added[group] = value
                        self.addIPRangeToGroup(group, value, mask)
            elif key.startswith('REMOVERANGE_'):
                group = key[12:]
                values = REQUEST.form[key].split(';')
                for value in values:
                    if value.strip():
                        mask = None
                        val_mask = value.split('/')
                        if len(val_mask)==2 and val_mask[1]:
                            mask = val_mask[1]
                            value = val_mask[0]            
                        removed[group] = value
                        self.removeIPRangeFromGroup(group, value, mask)
        return REQUEST.RESPONSE.redirect(self.absolute_url() +
                                         '/manage_IPRangeGroupMappings')                      
                

    security.declareProtected( ManageGroups, 'manage_IPRangeGroupMappings' )
    manage_IPRangeGroupMappings = PageTemplateFile(
        os.path.join(_wwwdir, 'manage_ip_range_group_mappings'), 
        globals(),
        __name__='manage_IPRangeGroupMappings')
        
    manage_options =  ( { 'label' : 'IP to Group Mapping'
                        , 'action' : 'manage_IPRangeGroupMappings'
                        }
                      ,
                      ) + BasePlugin.manage_options + Cacheable.manage_options
        
classImplements(GroupByIPPlugin, IGroupsPlugin)
InitializeClass(GroupByIPPlugin)